/*
 * CSE 142 Spring 2001
 * Homework 4, Part A
 * Skeleton code by Luke Meyers, Hannah Tang, David Dunham
 *
 * hw4.c
 * Main file for the Homework 4 project.  This file contains all the functions
 * you should need to change for Homework 4.  You shouldn't need to change the
 * stuff in the other files (tron.h, tronDraw.c, tronMove.c), but you should
 * look them over because they contain symbolic constants and functions that
 * you'll need to make use of.
 *
 * Author:
 * Student ID:
 * Quiz Section:
 * TA:
 * Program Description:
 *
 */
 
#include "GP142.h"
#include <stdio.h>
#include <math.h>
#include <time.h>

/* tron.h is a header file that defines symbolic constants used in more than one
 * .c file, and prototypes of functions that are called from a file other than
 * the one they are defined in. */
#include "tron.h"

/******************************************************************************
 * Symbolic constants.  There are a lot more of these in tron.h.  If you need
 * to make new ones, they should probably go here unless they're needed by the
 * drawing or movement code.
 ******************************************************************************/

/* String constants */
#define STR_LEVEL_START		"Press any key to start."
#define STR_PLAYER_DIED		"You died.\n\nPress any key to continue."
#define STR_FLASH_SCREEN	"Use numeric keypad to control\nplayer, and space bar to shoot."
#define STR_LEVEL_END		"Congratulations!\n\nHit space to play the next level."
#define STR_GAME_OVER		"GAME OVER"

/* Level start settings, and game mechanics */
#define PLAYER_START_X		0
#define PLAYER_START_Y		0
#define PLAYER_START_FACING	(-LEFT)
#define INITIAL_LIVES		3
#define MIN_ENEMY_DIST		70	/* How far away enemies must be from player start position */
#define ENEMY_SCORE			50
#define LEVEL_END_SCORE		200

/******************************************************************************
 * Function prototypes.  See the function definitions for 
 * descriptions of what they do.
 ******************************************************************************/
void startGame(int *gameStatus, int *level, int *numLives, int *score,
			   int *playerX, int *playerY, int *playerFacing,
			   int *enemyX,  int *enemyY, int *enemyStatus,
			   int *bulletX, int *bulletY, int *bulletStatus, int *bulletFacing);
void endGame(int *gameStatus);
void startLevel(int *gameStatus, int level, 
				int *playerX, int *playerY, int *playerFacing,
			    int *enemyX,  int *enemyY, int *enemyStatus,
			    int *bulletX, int *bulletY, int *bulletStatus, int *bulletFacing);
void endLevel(int *gameStatus, int *level, int *score);
void killPlayer(int *gameStatus, int *numLives);
void createEnemy(int *enemyX, int *enemyY, int *enemyStatus);
void killEnemy(int level, int *bulletStatus, int *enemyStatus, int *score);
int enemiesLeft(int enemyStatus);

void drawScreen(int gameStatus, int level, int numLives, int score,
				int playerX, int playerY, int playerFacing,
				int enemyX, int enemyY, int enemyStatus,
				int bulletX, int bulletY, int bulletStatus, int bulletFacing);
void drawInGameScreen(int level, int numLives, int score,
					  int playerX, int playerY, int playerFacing,
					  int enemyX, int enemyY, int enemyStatus,
					  int bulletX, int bulletY, int bulletStatus, int bulletSpacing);

void updateScreen(int *gameStatus, int *level, int *numLives, int *score,
				  int *playerX, int *playerY, int *playerFacing,
				  int *enemyX,  int *enemyY, int *enemyStatus,
				  int *bulletX, int *bulletY, int *bulletStatus, int *bulletFacing);
void handleInput(int *gameStatus, int *level, int *numLives, int *score, char keyPressed, 
				 int *playerX, int *playerY, int *playerFacing,
				 int *enemyX,  int *enemyY, int *enemyStatus,
				 int *bulletX, int *bulletY, int *bulletStatus, int *bulletFacing);
void shootBullet(int *bulletX, int *bulletY, int *bulletStatus, int *bulletFacing, 
				 int playerX, int playerY, int playerFacing);

/******************************************************************************
 *
 * The Main Function
 * ----------------
 *
 * The main function takes care of initialization and cleanup, and the high-level
 * control of all actions in the program, but most of the detail work is 
 * relegated to other functions, so that the structure of main stays clear and 
 * simple.
 *
 */
int main(void)
{                                
	/* Event loop management */
    int quit = FALSE;       /* Has user signaled quit yet?                      */
	int nextEvent;			/* The next event in the event loop.  				*/
    int mouseX, mouseY;
    char keyPressed;

	/* Game state */
	int numLives,
		level,
		score,
		gameStatus = GAME_FLASH_SCREEN;

	/* Objects' state */
	int playerX,
		playerY,
		playerFacing;		/* Direction in which the player is facing. */
	int enemyX,
		enemyY,
		enemyStatus;
	int bulletX,
		bulletY,
		bulletStatus,
		bulletFacing;		/* Direction the bullet is pointed in. */

    GP142_open();           /* Open and initialize the GP142 Graphics Window    */
    GP142_logging(LOG_OFF); /* logging is useful to students during debugging,  */
                            /* but annoying during the demo; turn it off.       */
	GP142_animate(ANI_RUN); /* Start animation */

	srand(time(NULL));		/* Seed the random number generator */


    /*
     * The main event loop:
     * --- ---- ----- -----
     * Wait for the next user action, such as a mouse click or keyboard
     * keypress, or a periodic update event (allowing us to display the
     * next frame of an animation) and call the appropriate function to
     * handle it.  Repeat this until the user signals "quit".
     */
     
    while (!quit) 
	{
		/* Find out which event has happened. */
		nextEvent = GP142_await_event(&mouseX, &mouseY, &keyPressed);

        switch (nextEvent) 
		{
            case GP142_QUIT:
                quit = TRUE;   /* set flag to terminate loop                */
                break;
   
            case GP142_MOUSE:
                break;
               
            case GP142_KBD:
				if (keyPressed == 'q' || keyPressed == 'Q')
				{
					quit = TRUE;
				}
				else 
				{
					/* The user has typed in some keyboard input.  Send it,
					 * along with a bunch of useful pointers, off to the
					 * function that handles keyboard input. */
					handleInput(&gameStatus, &level, &numLives, &score, keyPressed, 
								&playerX, &playerY, &playerFacing,
								&enemyX, &enemyY, &enemyStatus, 
								&bulletX, &bulletY, &bulletStatus, &bulletFacing);
				}
                break;

            case GP142_PERIODIC:
				/* GP142_PERIODIC is an event that gets triggered roughly every
				 * tenth of a second.  This is where one does things like drawing
				 * the next frame of animation, or updating the program state by
				 * one "tick." */
				if (gameStatus == GAME_PLAYING)
				{
					updateScreen(&gameStatus, &level, &numLives, &score, 
						&playerX, &playerY, &playerFacing, 
						&enemyX, &enemyY, &enemyStatus, 
						&bulletX, &bulletY, &bulletStatus, &bulletFacing);
				}
				drawScreen(gameStatus, level, numLives, score, 
					playerX, playerY, playerFacing, 
					enemyX, enemyY, enemyStatus, 
					bulletX, bulletY, bulletStatus, bulletFacing);
                break;

            default:
                break;
        }
    }
        
    GP142_close();          /* Clean up and close graphics window           */
    
    return 0;
}

/******************************************************************************
 * startGame
 * 
 * Set up everything necessary to get the game started, then start the first
 * level.
 *
 * Parameters: pointers to the various object attributes and game state data
 */
void startGame(int *gameStatus, int *level, int *numLives, int *score,
			   int *playerX, int *playerY, int *playerFacing,
			   int *enemyX,  int *enemyY, int *enemyStatus,
			   int *bulletX, int *bulletY, int *bulletStatus, int *bulletFacing)
{
	*gameStatus = GAME_LEVEL_START;
	*level = 1;
	*score = 0;
	*numLives = INITIAL_LIVES;
	startLevel(gameStatus, *level, playerX, playerY, playerFacing, enemyX, enemyY, 
		enemyStatus, bulletX, bulletY, bulletStatus, bulletFacing);
}

/******************************************************************************
 * endGame
 * 
 * End the game, and set it to show the "Game Over" screen.
 *
 * Parameters:	gameStatus: pointer to game status so it can be updated.
 */
void endGame(int *gameStatus)
{
	*gameStatus = GAME_OVER;
}

/******************************************************************************
 * startLevel
 *
 * Start a level with the given level number.
 *
 * Parameters:	The level number, plus pointers to the various game object
 *				attributes and state data.
 */
void startLevel(int *gameStatus, int level, 
				int *playerX, int *playerY, int *playerFacing,
			    int *enemyX,  int *enemyY, int *enemyStatus,
			    int *bulletX, int *bulletY, int *bulletStatus, int *bulletFacing)
{
	/* TODO:
	 * Initialize the player's position and facing.  Create an enemy, and set the
	 * bullet's status appropriately. */
}

/******************************************************************************
 * endLevel
 *
 * End the current level and set it to show the end-level screen.  Also,
 * endLevel is responsible for incrementing the level number.
 *
 * Parameters:	pointers to various game state variables that need to be updated.
 */
void endLevel(int *gameStatus, int *level, int *score)
{
	*gameStatus = GAME_LEVEL_END;
	*score += LEVEL_END_SCORE;
	*level += 1;
}

/******************************************************************************
 * killPlayer
 *
 * The player has been hit by an enemy, so cause it to die and start a new life
 * if one is available.  If not, end the game.
 * 
 * Parameters:	gameStatus: pointer to the game status, which should be updated
 *				numLives: pointer to player's lives, which should be decreased by 1.
 */
void killPlayer(int *gameStatus, int *numLives)
{
	/* TODO: Handle player death. */
}

/******************************************************************************
 * createEnemy
 *
 * Initialize one enemy's status and position.
 *
 * Parameters: pointers to the attributes of a single enemy.
 */
void createEnemy(int *enemyX, int *enemyY, int *enemyStatus)
{
	*enemyStatus = ENEMY_ALIVE;

	/* We want to make sure the enemy starts at least a certain distance away
	 * from the player, to make the game fair.  So, we loop until the values
	 * are acceptable.  Note that we start out with "bad" values, so that 
	 * the loop always executes at least once. */
	*enemyX = PLAYER_START_X;
	*enemyY = PLAYER_START_Y;
	while(abs(*enemyX - PLAYER_START_X) < MIN_ENEMY_DIST
		  && abs(*enemyY - PLAYER_START_Y) < MIN_ENEMY_DIST)
	{
		*enemyX = (rand() % (GP142_XMAX*2)) - GP142_XMAX;
		*enemyY = (rand() % (GP142_YMAX*2)) - GP142_YMAX;
	}
}

/******************************************************************************
 * killEnemy
 *
 * An enemy has been hit by a player's bullet, so cause it to die and do all
 * the related stuff: increase score, remove enemy from screen, etc.
 *
 * Parameters:	level: The current level of the game
 *				score: pointer to the player's score
 *				bulletStatus: pointer to the status of the bullet
 *				enemyStatus: pointer to the status of the enemy
 */
void killEnemy(int level, int *score, int *bulletStatus, int *enemyStatus)
{
	/* TODO:
	 * This function should be called when an enemy is shot.  It should make
	 * the enemy be dead, and increase the player's score.  Also make the
	 * bullet disappear, not keep going! */
}

/******************************************************************************
 * enemiesLeft
 *
 * Are there any enemies left alive on the current level?
 *
 * Parameters:	enemyStatus: The status of the enemy.
 */
int enemiesLeft(int enemyStatus)
{
	if (enemyStatus == ENEMY_ALIVE)
	{
		return TRUE;
	}
	return FALSE;
}

/******************************************************************************
 * drawScreen
 *
 * Draw the entire screen.  Other drawing functions should be called from
 * inside this one.
 * 
 * Parameters: A whole bunch of data about game state and object attributes.
 *				Note that *none* of these are pointers (output parameters)!
 */
void drawScreen(int gameStatus, int level, int numLives, int score,
				int playerX, int playerY, int playerFacing, 
				int enemyX, int enemyY, int enemyStatus,
				int bulletX, int bulletY, int bulletStatus, int bulletFacing)
{
	GP142_clear();

	if (gameStatus != GAME_FLASH_SCREEN)
	{
		/* Draw the screen of the game as it appears in play. */
		drawInGameScreen(level, numLives, score,
						 playerX, playerY, playerFacing,
						 enemyX, enemyY, enemyStatus,
						 bulletX, bulletY, bulletStatus, bulletFacing);
	}

	switch(gameStatus)
	{
	case GAME_FLASH_SCREEN:
		/* Show a flashy title screen. */

		/* TODO: Make the title screen show the right info 
		 * (author name, game name, instructions). */

		/* Draw background */
		drawBackground();
		drawMessage(STR_FLASH_SCREEN);
		break;

	case GAME_PLAYING:
		break;

	case GAME_LEVEL_START:
		drawMessage(STR_LEVEL_START);
		break;

	case GAME_PLAYER_DIED:
		drawMessage(STR_PLAYER_DIED);
		break;

	case GAME_LEVEL_END:
		drawMessage(STR_LEVEL_END);
		break;
		
	case GAME_OVER:
		drawMessage(STR_GAME_OVER);
		break;
	}
}

/******************************************************************************
 * drawInGameScreen
 *
 * Draw all the stuff that appears on the game screen when the game is
 * actually in progress.
 *
 * Parameters:	level, numLives, score: The appropriate game state data
 *				other parameters: Various object attributes.
 */
void drawInGameScreen(int level, int numLives, int score,
					  int playerX, int playerY, int playerFacing,
					  int enemyX, int enemyY, int enemyStatus,
					  int bulletX, int bulletY, int bulletStatus, int bulletFacing)
{
	drawBackground();
	drawBanner(level, numLives, score);

	/* Draw game objects */
	drawPlayer(playerX, playerY, playerFacing);
	drawEnemy(enemyX, enemyY, enemyStatus);
	drawBullet(bulletX, bulletY, bulletStatus, bulletFacing);
}

/******************************************************************************
 * updateScreen
 *
 * Perform one "frame" of animation by updating the positions of everything
 * on the screen.  This does not include actually drawing those things, just
 * updating their position variables.
 *
 * Parameters: the number of the current level, plus pointers to the current 
 *				positions of everything and various status stuff.
 */
void updateScreen(int *gameStatus, int *level, int *numLives, int *score,
				  int *playerX, int *playerY, int *playerFacing,
				  int *enemyX,  int *enemyY, int *enemyStatus,
				  int *bulletX, int *bulletY, int *bulletStatus, int *bulletFacing)
{
	/* We need the bullet's new and old position in order to do collision
	 * detection properly. */
	int oldBulletX = *bulletX,
		oldBulletY = *bulletY;
	movePlayer(playerX, playerY, *playerFacing);
	moveEnemy(enemyX, enemyY, *playerX, *playerY);
	moveBullet(bulletX, bulletY, bulletStatus, *bulletFacing);

	/***********************
	 * Collision detection *
	 ***********************/

	/* Player hit by enemy? */
	if (rectanglesOverlap(*playerX, *playerY, PLAYER_WIDTH, PLAYER_HEIGHT,
						*enemyX, *enemyY, ENEMY_WIDTH, ENEMY_HEIGHT)
		&& *enemyStatus == ENEMY_ALIVE)
	{
		killPlayer(gameStatus, numLives);
		if (*numLives < 0)
		{
			endGame(gameStatus);
		}
		else
		{
			/* Start the level over with the remaining enemies. */
			startLevel(gameStatus, *level, 
				playerX, playerY, playerFacing, 
				enemyX, enemyY, enemyStatus, 
				bulletX, bulletY, bulletStatus, bulletFacing);
		}
	}


	/* Enemy hit by bullet? */
	if (*bulletStatus == BULLET_ALIVE)
	{
		/* TODO: Determine whether the bullet has hit an enemy.  If so, kill the
		 * enemy.  If all enemies are dead, end the level. */
	}
}


/******************************************************************************
 * handleInput
 *
 * Perform the appropriate action, based on which key was pressed.
 * This function needs pointers to a lot of stuff, because it calls various
 * other functions that need them.
 *
 * Parameters:	gameStatus, level, numLives, score: pointers to game state data
 *				keyPressed: the character typed on the keyboard
 *				other paramerers: Pointers to various object attributes.
 */
void handleInput(int *gameStatus, int *level, int *numLives, int *score, char keyPressed, 
				 int *playerX, int *playerY, int *playerFacing,
				 int *enemyX,  int *enemyY, int *enemyStatus,
				 int *bulletX, int *bulletY, int *bulletStatus, int *bulletFacing)
{

	switch(*gameStatus)
	{
	case GAME_FLASH_SCREEN:
		startGame(gameStatus, level, numLives, score,
			playerX, playerY, playerFacing,
			enemyX, enemyY, enemyStatus,
			bulletX, bulletY, bulletStatus, bulletFacing);
		break;
	case GAME_LEVEL_START:
		*gameStatus = GAME_PLAYING;
		break;
	case GAME_PLAYER_DIED:
		*gameStatus = GAME_PLAYING;
		break;
	case GAME_LEVEL_END:
		if (keyPressed == ' ') /* Hit space to continue.  Should this be a symbolic constant? */
		{
			startLevel(gameStatus, *level, 
					   playerX, playerY, playerFacing,
					   enemyX, enemyY, enemyStatus,
					   bulletX, bulletY, bulletStatus, bulletFacing);
			*gameStatus = GAME_LEVEL_START;
		}
		break;
	case GAME_OVER:
		*gameStatus = GAME_FLASH_SCREEN;
		break;
	case GAME_PLAYING:

		/* Nested switch statement!  The cases in here are not seen by the outer switch
		 * statement!  Be careful! */
		switch(keyPressed)
		{
		/* TODO:
		 * Add keys (space bar and zero) for shooting a bullet. */

		/* Movement keys. */
		case '1':
			*playerFacing = DOWN_LEFT;
			break;
		case '2':
			*playerFacing = DOWN;
			break;
		case '3':
			*playerFacing = DOWN_RIGHT;
			break;
		case '4':
			*playerFacing = LEFT;
			break;
		case '5':
			/* Stop the player.  Remember, this works by negating playerFacing.  If the
			 * player is already stopped, this sets it going again. */
			*playerFacing = -(*playerFacing);
			break;
		case '6':
			*playerFacing = RIGHT;
			break;
		case '7':
			*playerFacing = UP_LEFT;
			break;
		case '8':
			*playerFacing = UP;
			break;
		case '9':
			*playerFacing = UP_RIGHT;
			break;
		default:
			break;
		}
		break;
	}


}

/******************************************************************************
 * shootBullet
 *
 * Shoot a bullet in the indicated direction.
 *
 * Parameters: bulletX and bulletY: pointers to the bullet's position
 *			   direction: the direction to shoot in.
 */
void shootBullet(int *bulletX, int *bulletY, int *bulletStatus, int *bulletFacing, 
				 int playerX, int playerY, int playerFacing)
{
	if (*bulletStatus == BULLET_DEAD)
	{
		*bulletStatus = BULLET_ALIVE;
		*bulletFacing = abs(playerFacing); /* Use abs because we don't want the bullet
											* to be stopped midair! */
		*bulletX = playerX + PLAYER_WIDTH/2 - BULLET_WIDTH/2;
		*bulletY = playerY + PLAYER_HEIGHT/2 - BULLET_HEIGHT/2;
	}
}